import signal
import time
import re
import random
import os
import shutil
import numpy as np
import libpsf
import tarfile
import threading
import string
import concurrent.futures
import subprocess

from io import BytesIO
from hashlib import sha1 as filehashfn
from textwrap import wrap as textwrap

from pexpect.replwrap import REPLWrapper

import waverunner

from scipy.interpolate import interp1d

from multiprocessing.managers import BaseManager
from multiprocessing import Lock

#from .spectre import get_REPLWrapper_in_another_process, get_REPLWrapper_in_this_process
from .spectre.skill_interface import Accelerated_scl_interface

from . import default_spectre_binary, default_rcfile, default_spectre_env
from . import pyspectre_path, pyspectre_ckt_dir, pyspectre_sim_dir
from .spectre.replwrap_server import REPLWrapper_client

_spectre_managed_models = False
_spectre_env_flag = False
_rcfile = None
_spectre_env = None
_spectre_binary = default_spectre_binary

float_fmt_string = '{:+13.6e}'
local_sim_db_relpath = 'PYSPECTRE_AHDL_SIM'
local_ship_db_relpath = 'PYSPECTRE_AHDL_SHIP'


class LockManager(BaseManager):
    pass


try:
    LockManager.register('global_lock')
    global_manager = LockManager(address=('', 11223), authkey=b'iamjustalittlepyspectre')
    global_manager.connect()

except BaseException as e:
    # there is no lock server... start one:
    global_lock = Lock()
    LockManager.register('global_lock', callable=lambda: global_lock)
    global_manager = LockManager(address=('', 11223), authkey=b'iamjustalittlepyspectre')
    global_manager.start()

# def get_rcfile():
#     return _rcfile
#
#
# def get_spectre_env():
#     return _spectre_env
#
#
# def get_spectre_binary():
#     return _spectre_binary


def set_rcfile(filepath=None):
    global _rcfile

    if filepath is None:
        filepath = default_rcfile

    filespec = os.path.expandvars(os.path.expanduser(filepath))

    if os.path.isfile(filespec):
        _rcfile = filespec


def set_spectre_env(env=None):
    global _spectre_env

    if env is None:
        _spectre_env = default_spectre_env
    else:
        assert isinstance(env, (tuple, list))
        assert all([isinstance(item, tuple) for item in env])
        assert all([isinstance(item, str) for pairs in env for item in pairs])
        _spectre_env = env


def set_spectre_binary(exepath=None):
    global _spectre_binary

    if exepath is None:
        _spectre_binary = default_spectre_binary
    else:
        specpath = os.path.expandvars(os.path.expanduser(exepath))
        assert os.path.isfile(specpath)
        _spectre_binary = specpath


def format_float(float):
    return float_fmt_string.format(float)


def format_vect(vect):
    return ' '.join([float_fmt_string.format(v) for v in vect.T.flatten()])


def _digest_file(filepath):
    return filehashfn(open(filepath, mode='rb').read().replace(' '.encode(), ''.encode())).hexdigest()


def _digest_string(string):
    return filehashfn(string.encode()).hexdigest()


def _get_patterns_for_internal_replacement(path_representation):
    return [r'^.\d*({})_.*[.]dep$'.format(path_representation)]


def hash_test(stimuli, time_vec):
    return str(filehashfn(
        str(tuple(np.concatenate([time_vec.ravel(), stimuli.ravel()]))).encode('utf-8')
    ).hexdigest()[:16])


def hash_tests(stimsets, time_vect):
    ids = []
    for tt in stimsets:
        ids.append(hash_test(tt, time_vect))
    return ids


class PyspectreUnresolvableError(ValueError):
    pass


class PwlInput(object):
    def __init__(self, name, max_n_points, parent=None):
        self.max_n_points = max_n_points
        self.name = name
        self.cname = '.'.join(name, parent) if parent else name
        self.default_value = None
        self.recent_value = None


class Parameter(object):
    def __init__(self, name, parent=None, ):
        if parent is not None:
            name = '.'.join(name, parent)
        self.name = name
        self.default_value = None
        self.recent_value = None


class Circuit(object):
    def __init__(self,
                 netlist=None,
                 simdirectory=None,
                 analysis_name='tran',
                 spectre_args=None,
                 pwl_inputs=None,
                 parameters=None,
                 ):

        # settle netlist file
        if not os.path.isfile(netlist):
            # print('netlist \"{}\" is not file'.format(netlist))
            # print('cwd: \"{}\"'.format(os.getcwd()))
            if not os.path.isdir(simdirectory):
                # print('simdirectory \"{}\" is not directory'.format(simdirectory))
                # print('cwd: \"{}\"'.format(os.getcwd()))
                raise IOError(
                    'unable to find spice netlist: \"{}\" simdirectory: \"{}\"'.format(netlist, simdirectory))
            netlist = os.path.join(simdirectory, netlist)
            if not os.path.isfile(netlist):
                # print('netlist appended.  still not file: \"{}\"'.format(netlist))
                # print('cwd: \"{}\"'.format(os.getcwd()))
                raise IOError(
                    'unable to find spice netlist: netlist \"{}\" simdirectory: \"{}\"'.format(netlist, simdirectory))

        # settle simdir
        if simdirectory is not None and not os.path.isdir(simdirectory):
            simdirectory = os.path.dirname(netlist)

        if spectre_args is not None:
            if isinstance(spectre_args, list):
                spectre_args = ' '.join(spectre_args)
            elif isinstance(spectre_args, str):
                spectre_args = spectre_args

        if pwl_inputs is not None:
            if isinstance(pwl_inputs, PwlInput):
                pwl_inputs = tuple(pwl_inputs.cname, pwl_inputs)
            else:
                assert isinstance(pwl_inputs, (list, tuple))
                assert all([isinstance(item, PwlInput) for item in pwl_inputs])
                pwl_inputs = tuple(pwl_inputs)

        if parameters is not None:
            if isinstance(parameters, Parameter):
                parameters = tuple(parameters)
            else:
                assert isinstance(parameters, (list, tuple))
                assert all([isinstance(item, Parameter) for item in parameters])
                parameters = tuple(parameters)

        # build command
        # self.resolve_spectre_modes(spectre_output_mode, spectre_reader)

        self.spicefilepath = os.path.abspath(netlist)
        self.simdir = simdirectory
        self.analysis_name = analysis_name
        self.spectre_args = spectre_args
        self.pwl_inputs = pwl_inputs
        self.parameters = parameters
        self.tarball = None
        self.id = str(hash(self))

    # def resolve_spectre_modes(self, spectre_output_mode, spectre_reader):
    #     if spectre_output_mode is None and spectre_reader is None:
    #         self.spectre_output_mode = 'psfbin'
    #         self.spectre_reader = 'libpsf'
    #
    #     elif spectre_reader is not None and spectre_output_mode is None:
    #         assert spectre_reader in avail_readers
    #         self.spectre_reader = spectre_reader
    #         if spectre_reader == 'decida':
    #             self.spectre_output_mode = 'nutascii'
    #         elif spectre_reader == 'libpsf':
    #             self.spectre_output_mode = 'psfbin'
    #         else:
    #             raise NotImplementedError
    #
    #     elif spectre_output_mode is not None and spectre_reader is None:
    #         assert spectre_output_mode in avail_output_modes
    #         self.spectre_output_mode = spectre_output_mode
    #         if spectre_output_mode == 'ascii':
    #             self.spectre_reader = 'nutascii'
    #         elif spectre_output_mode == 'psfbin':
    #             self.spectre_reader = 'libpsf'
    #         else:
    #             raise NotImplementedError
    #
    #     else:
    #         assert spectre_reader in avail_readers
    #         self.spectre_reader = spectre_reader
    #         assert spectre_output_mode in avail_output_modes
    #         self.spectre_output_mode = spectre_output_mode

    def pack_up(self):
        self.tarball = self._make_tarball()

    def unpack(self, path):
        if not path:
            path = os.getcwd()
        self.tarball.seek(0)
        tb = tarfile.open(fileobj=self.tarball, mode='r:bz2')
        tb.extractall(path=path)
        tb.close()

    def _make_tarball(self):
        tarball = BytesIO()
        tb = tarfile.open(fileobj=tarball, mode='w:bz2')
        for item in os.listdir(self.simdir):
            tb.add(os.path.join(self.simdir, item), arcname=item)
        tb.close()
        return tarball

    def __hash__(self):
        return int(_digest_string(''.join(self._hash_files())), 16)

    def _hash_files(self):
        file_hashes = {}
        file_hashes.update({self.spicefilepath: _digest_file(self.spicefilepath)})

        if self.simdir is not None:
            for bdir, sdirs, files in os.walk(self.simdir):
                for ff in files:
                    filepath = os.path.join(bdir, ff)
                    file_hashes.update({filepath: _digest_file(filepath)})

        hashes = list(file_hashes.values())
        hashes.sort()
        return hashes

    def run_batch(self, *args, **kwargs):
        return run_batch(self, *args, **kwargs)

    def start_pyspectre(self):
        return Pyspectre(circuit=self, asynch_repl=False, logging=False, verbosity=0)


def do_batch_on_runner(runner, list_of_stimuli, list_of_times, signal_names, resample, verbosity):
    return runner.run_batch(
        list_of_stimuli=list_of_stimuli,
        list_of_times=list_of_times,
        signal_names=signal_names,
        resample=resample,
        verbosity=verbosity,
    )


def do_batch_here(circuit, list_of_stimuli, list_of_times, signal_names, resample, verbosity, logging, asynch_repl, shutdown_event=None):
    print('starting pyspectre...') if verbosity >=4 else None
    with Pyspectre(circuit=circuit, logging=logging, verbosity=verbosity, asynch_repl=asynch_repl, shutdown_event=shutdown_event) as pyspectre:
        return do_batch_on_runner(pyspectre, list_of_stimuli, list_of_times, signal_names, resample, verbosity)


def run_batch(circuit, list_of_stimuli, list_of_times, signal_names='all', resample=False, parallel=False, n_workers=0,
              max_jobsize=10, verbosity=1, local=True, timeout=600, logging=False, asynch_repl='thread'):
    if not signal_names or signal_names is None:
        signal_names = 'all'

    if not parallel:
        return do_batch_here(circuit, list_of_stimuli, list_of_times, signal_names, resample, verbosity, logging, asynch_repl)

    elif parallel:

        nsims = list_of_stimuli.shape[0]
        allinds = np.arange(0, nsims)
        np.random.shuffle(allinds)
        expected_sims_per_worker = int(np.ceil(nsims / n_workers))

        # resort by expected duration::::
        by_duration = np.argsort(list_of_times[allinds, -1]).reshape(2, -1).T
        by_duration[:, 1] = by_duration[:, 1][::-1]
        allinds = by_duration.flatten()

        if expected_sims_per_worker > max_jobsize:
            n_full_batches = int(np.ceil(nsims / max_jobsize))
            batch_inds = np.array_split(allinds, n_full_batches, axis=0)
        else:
            njobs = min([n_workers, nsims])
            batch_inds = np.array_split(allinds, njobs, axis=0)

        SIM_IDS = [hash((str(s), str(t))) for s, t in zip(list_of_stimuli, list_of_times)]
        JOB_IDS = [hash(str([SIM_IDS[ind] for ind in inds])) for inds in batch_inds]

        if local:

            response_times = [None]*nsims
            response_sigs = [None]*nsims
            print('submitting {} batches'.format(len(batch_inds))) if verbosity >= 1 else None
            otime = time.time()

            with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
                shutdown = threading.Event()
                try:
                    jobs = {executor.submit(do_batch_here,
                                            circuit=circuit,
                                            list_of_stimuli=list_of_stimuli[inds],
                                            list_of_times=list_of_times[inds],
                                            signal_names=signal_names,
                                            resample=resample,
                                            verbosity=verbosity,
                                            logging=logging,
                                            asynch_repl=asynch_repl,
                                            shutdown_event=shutdown,
                                            ): JOB_IDS[i] for i, inds in enumerate(batch_inds)}

                    for future in concurrent.futures.as_completed(jobs):
                        jid = jobs[future]
                        # print('[pyspectre] completed batch {} of {}'.format(JOB_IDS.index(jid)+1, len(JOB_IDS))) if verbosity >= 1 else None
                        print('[pyspectre] {} batches remaining'.format(sum([1 for j in jobs if not j.done() and not j.cancelled()]))) if verbosity >= 1 else None
                        jsiminds = batch_inds[JOB_IDS.index(jid)]
                        try:
                            jtimes, jresponses = future.result()
                        except Exception as exc:
                            print('%r generated an exception: %s' % (jid, exc))
                        else:
                            for rsimind, rtime, rresponse in zip(jsiminds, jtimes, jresponses):
                                response_times[rsimind] = rtime
                                response_sigs[rsimind] = rresponse

                except KeyboardInterrupt as e:
                    for future in jobs:
                        future.cancel()
                    shutdown.set()
                    executor.shutdown(wait=True)
                    raise e

            elapsed = time.time() - otime
            print('all batches took {:5.1f} s to complete'.format(elapsed))
            return response_times, response_sigs

        elif not local:

            queuer = waverunner.Waverunner(notify_interval=0, polling_interval=0, verbosity=0)
            queuer.start_service()

            jobids = []
            circuit.solver = None
            for inds in batch_inds:
                jobids.append(queuer.queue_job(
                    'run_batch',
                    kwargs={
                        'circuit': circuit,
                        'list_of_stimuli': list_of_stimuli[inds],
                        'list_of_times': list_of_times[inds],
                        'signal_names': signal_names,
                        'resample': resample,
                        'verbosity': verbosity,
                        'logging': logging,
                        'asynch_repl': asynch_repl,
                    },
                    timeout=timeout,
                    # completion_hook=completion_hook,
                ))

            print('waiting for up to {} seconds for all jobs to complete'.format(timeout)) if verbosity >= 1 else None
            queuer.block_for_queue(timeout=timeout)

            completed = queuer.get_completed_jobs(flush=False)
            results = []
            for id in jobids:
                if id in [j.id for j in completed]:

                    job = completed[[j.id for j in completed].index(id)]

                    if results is None:
                        results = job.result
                    else:
                        results = np.concatenate([results, job.result], axis=0)

            queuer.get_completed_jobs(flush=True)
            queuer.stop_service()
            return results


class Pyspectre(object):

    def __init__(self,
                 circuit=None,
                 netlist=None,
                 simdirectory=None,
                 analysis_name='tran',
                 spectreargs=None,
                 logging=False,
                 pwl_inputs=None,
                 asynch_repl='thread',
                 verbosity=1,
                 shutdown_event=None,
                 ):

        self.verbosity = verbosity

        if isinstance(threading.current_thread(), threading._MainThread):
            signal.signal(signal.SIGINT, self._signal_handler)
            signal.signal(signal.SIGABRT, self._signal_handler)
            signal.signal(signal.SIGTERM, self._signal_handler)

        if circuit is None:
            self.circuit = Circuit(
                netlist=netlist,
                simdirectory=simdirectory,
                analysis_name=analysis_name,
                spectre_args=spectreargs,
                pwl_inputs=pwl_inputs
            )
        else:
            self.circuit = circuit

        self.spectre_args = circuit.spectre_args
        self.spicefilepath = circuit.spicefilepath
        self.simdir = circuit.simdir
        self.analysis_name = circuit.analysis_name
        self.logging = logging
        self.pwl_inputs = circuit.pwl_inputs
        self.asynch_repl = asynch_repl
        self.shutdown_event = shutdown_event

        self.original_simdir = None
        self.analysis_type = None
        self.dcsolution = None
        self.working_path = None
        self.results_dir = None
        self.quit = False
        self.spectre = None
        self.interface = None
        self.fresh = True
        self.pid = None
        self.commandct = None
        self.spectre = None
        self.cmd = None
        self.reader = 'libpsf'

        self._start_binary()

    def __del__(self):
        self._cleanup()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._cleanup()
        del self
        return

    def _signal_handler(self, signal, frame):
        self._cleanup()
        raise KeyboardInterrupt

    def _cleanup(self):
        self.quit = True
        try:
            self.interface.quit()
        except:
            try:
                self.spectre.child.terminate()
            except:
                try:
                    self.spectre.child.kill(15)
                except:
                    try:
                        self.spectre.child.kill(9)
                    except:
                        try:
                            os.kill(self.pid, 9)
                        except:
                            pass

        try:
            self.spectre.child.join()
        except:
            pass

        try:
            self.spectre.child.wait()
        except:
            pass

        try:
            shutil.rmtree(self.working_path)
        except:
            pass

    def _purge_stray_workingdirs(self):

        # check for stray data
        rundirs = os.listdir(pyspectre_sim_dir)
        active = [line.split() for line in subprocess.check_output(['bash', '-c', 'ps -ef | grep pyspectre']).decode().split('\n') if len(line.split()) == 22]
        active = [(os.path.split(item[9])[-1], item[1]) for item in active]
        for dir in rundirs:
            if dir not in [item[0] for item in active]:
                try:
                    print('removing stray simulation: {}'.format(dir)) if self.verbosity >= 4 else None
                    shutil.rmtree(os.path.join(pyspectre_sim_dir, dir))
                except:
                    pass

    def _start_binary(self):
        global _spectre_env_flag

        if self.spectre is not None and self.spectre.child.isalive:
            return None
        if self.quit:
            return None

        print('creating basic pyspectre dirs') if self.verbosity >=4 else None
        # create/set shared working paths:
        os.makedirs(pyspectre_path, exist_ok=True)
        # os.makedirs(pyspectre_ahdl_lib, exist_ok=True)  # moving libraries to inside circuit #
        os.makedirs(pyspectre_sim_dir, exist_ok=True)
        os.makedirs(pyspectre_ckt_dir, exist_ok=True)



        # place circuit dir in cache
        cache_target = os.path.join(pyspectre_ckt_dir, self.circuit.id)
        if self.circuit.id not in os.listdir(pyspectre_ckt_dir):
            print('trying to get lock to cache circuit') if self.verbosity >= 4 else None

            # cache miss... get a lock:
            global_manager.global_lock().acquire(timeout=120)
            if self.circuit.id in os.listdir(pyspectre_ckt_dir):
                # somebody else put it there...
                global_manager.global_lock().release()
            else:
                # it's still not there... cache it
                print('caching circuit') if self.verbosity >= 4 else None

                if self.circuit.tarball is not None:
                    # unpack circuit in correct cache directory
                    self.original_simdir = self.simdir
                    self.circuit.unpack(cache_target)

                else:
                    if os.path.isdir(self.simdir):
                        # copy path to correct cache directory
                        shutil.copy(self.simdir, pyspectre_ckt_dir)
                        shutil.move(os.path.join(pyspectre_ckt_dir, os.path.split(self.simdir)[-1]),
                                    cache_target)

                    # copy spicefile
                    os.makedirs(cache_target, exist_ok=True)
                    shutil.copy(self.spicefilepath, cache_target)
                global_manager.global_lock().release()

        # setup working_path
        print('creating simulation dir') if self.verbosity >= 4 else None
        self.simid = ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.ascii_lowercase + string.digits)
                    for _ in range(10))

        self.working_path = os.path.join(pyspectre_sim_dir, self.simid)
        if os.path.isdir(self.working_path):
            os.removedirs(self.working_path)
        os.makedirs(self.working_path)
        self.results_dir = os.path.join(self.working_path, 'psf')
        os.makedirs(self.results_dir, exist_ok=False)

        # hardlink everything into sim-dir except for compiled AHDL objects (they have to be different)
        print('populating simulation dir') if self.verbosity >= 4 else None
        items = [os.path.join(cache_target, item) for item in os.listdir(cache_target)]
        for item in items:
            if os.path.split(item)[-1] == local_ship_db_relpath:
                shutil.copytree(item, os.path.join(self.working_path, local_ship_db_relpath))
            elif os.path.isdir(item):
                # shutil.copytree(item, os.path.join(self.working_path, os.path.split(item)[-1]), copy_function=os.link)
                ## above not py2.7 compatible
                os.system('cp -Rl {} {}'.format(item, os.path.join(self.working_path, os.path.split(item)[-1])))
            else:
                os.link(item, os.path.join(self.working_path, os.path.split(item)[-1]))

        self.original_simdir = self.simdir
        self.simdir = self.working_path

        if _spectre_managed_models:
            self.spicefilepath = os.path.join(cache_target, os.path.split(self.spicefilepath)[-1])
        else:
            self.spicefilepath = os.path.join(self.working_path, os.path.split(self.spicefilepath)[-1])

        # if you need to insure that there's enough memory for long waveforms:
        if self.pwl_inputs is not None:
            print('extending pwl definitions') if self.verbosity >= 4 else None
            self._extend_pwl_defs_in_netlist()

        ahdl_sim_local = os.path.join(self.working_path, local_sim_db_relpath)
        ahdl_ship_local = os.path.join(self.working_path, local_ship_db_relpath)
        os.makedirs(ahdl_sim_local, exist_ok=True)
        os.makedirs(ahdl_ship_local, exist_ok=True)

        # add special environment vars, if needed
        if _spectre_env is not None and not _spectre_env_flag:
            """
            don't use dict here, becuause we're sensitive to order... must execute in order
            """
            for (var, value) in _spectre_env:
                val = os.path.expandvars(value)
                if '$' in val:
                    val = ':'.join([item for item in val.split(':') if '$' not in item])
                print('{}={}'.format(var, val))
                os.environ[var] = val
            _spectre_env_flag = True

        # build up actual command-line call
        if _rcfile is not None:
            invocation = 'bash --noprofile --norc -ic \'stty -icanon && source {} && '.format(_rcfile)
        else:
            invocation = 'bash --noprofile --norc -ic \'stty -icanon && '

        cmd_args = [
            '-64',
            '+interactive=skill',
            '-log -note -info -debug -warn -error ++aps=moderate',
            '-format=psfbin',
            '-mt',
        ]

        if self.spectre_args:
            if isinstance(self.spectre_args, str):
                cmd_args.append(self.spectre_args)
            elif isinstance(self.spectre_args, (list, tuple)):
                cmd_args.extend(list(self.spectre_args))

        if self.logging:
            cmd_args.append('+log=\"spectre.log\"')

        self.cmd = ' '.join([
                'cd {}'.format(self.working_path),
                '&&',
                'CDS_AHDLCMI_SHIPDB_DIR={}'.format(os.path.join(self.working_path, local_ship_db_relpath)),
                'CDS_AHDLCMI_SIMDB_DIR={}'.format(os.path.join(self.working_path, local_sim_db_relpath)),
                'CDS_AHDLCMI_SHIPDB_COPY=YES',
                _spectre_binary,
                '-outdir {}'.format(self.working_path),
                ' '.join(cmd_args),
                # '-raw=\"raw\"',
                self.spicefilepath
            ])

        full_cmd = invocation + self.cmd + '\''

        try:
            ootime = time.time()
            if _spectre_managed_models:
                global_manager.global_lock().acquire(timeout=120)
                otime = time.time()
                print('spectre got lock in {:6.1f} ms'.format((otime - ootime) * 1000)) if self.verbosity>=2 else None

            else:
                self._load_relevant_lib_items()

            #### don't set variables in actual environment.... they get confused in parallelism!!!!!
            # os.environ['CDS_AHDLCMI_SIMDB_DIR'] = ahdl_sim_local
            # os.environ['CDS_AHDLCMI_SHIPDB_DIR'] = ahdl_ship_local
            # os.environ['CDS_AHDLCMI_SHIPDB_COPY'] = 'YES'

            if 'CDS_LIC_FILE' in os.environ:
                port, lic_srv = os.environ['CDS_LIC_FILE'].split('@')
                assert not os.system('ping -c 1 -W 1 {} > /dev/null'.format(lic_srv)), "couldn't reach cadence license server"
            else:
                print('cadence license server address not in shell env. trying \'spectre\' command anyway') if self.verbosity >=2 else None

            print('starting spectre binary') if self.verbosity >= 4 else None

            if self.asynch_repl in ['thread', 'threading']:

                self.spectre = REPLWrapper_client(
                    cmd_or_spawn=full_cmd,
                    orig_prompt=u'\n> ',
                    prompt_change=None,
                    server_process=False,
                    verbosity=self.verbosity,
                    shutdown_event=self.shutdown_event,
                )

            elif self.asynch_repl in ['process', 'processing', 'multiprocessing']:

                self.spectre = REPLWrapper_client(
                    cmd_or_spawn=full_cmd,
                    orig_prompt=u'\n> ',
                    prompt_change=None,
                    server_process=True,
                    verbosity=self.verbosity,
                    shutdown_event=self.shutdown_event,
                )

            else:
                try:
                    self.spectre = REPLWrapper(
                        cmd_or_spawn=full_cmd,
                        orig_prompt=u'\n> ',
                        prompt_change=None)

                    self.spectre.child.delayafterclose = 0.0
                    self.spectre.child.delaybeforesend = 0.0
                    self.spectre.child.delayafterread = 0.0
                    self.spectre.child.delayafterterminate = 0.0

                except Exception as e:
                    if hasattr(self, 'spectre') and hasattr(self.spectre, 'child') and self.spectre.child is not None:
                        print(self.spectre.child.before)
                    raise e

            self.fresh = False
            self.interface = Accelerated_scl_interface(self)
            self.spectre_pid = int(self.spectre.run_command('sclGetPid()'))
            self.bash_pid = self.spectre.child.pid
            self.commandct = 0
            self._resolve_analysis_type()

            if _spectre_managed_models:
                scllogfile = os.path.join(cache_target,
                                          '.' + os.path.split(self.spicefilepath)[1].split('.scs')[0] + '.scllog')

                if os.path.isfile(scllogfile):
                    os.link(scllogfile, os.path.join(self.working_path, 'scllog.scllog'))
                    os.remove(scllogfile)
                else:
                    raise RuntimeError('spectre did not create expected scllog file: {}'.format(scllogfile))

            startup_time = time.time() - ootime
            if startup_time < 1:
                print('started spectre in {:6.1f} ms'.format(startup_time * 1000)) if self.verbosity >= 1 else None
            elif startup_time >= 1:
                print('started spectre in {:6.1f} s'.format(startup_time)) if self.verbosity >= 1 else None
        finally:
            if _spectre_managed_models:
                global_manager.global_lock().release()
            else:
                self._cache_local_ship_db()

    def _cache_local_ship_db(self):
        local_ship_db = os.path.join(self.working_path, local_ship_db_relpath)
        local_lib_contents = os.listdir(local_ship_db)
        library_path = os.path.join(pyspectre_ckt_dir, self.circuit.id, local_ship_db_relpath)
        if not os.path.isdir(library_path):
            os.makedirs(library_path)

        def cache_item(s, d):

            if os.path.isfile(s):
                shutil.copy(s, d)

                # while not os.path.isfile(d):
                #     time.sleep(0.001)

                # with open(s, mode='rb') as ff:
                #     contents = ff.read()
                # contents = contents.replace(self.simid.encode(), self.circuit.id.encode())
                # with open(d, mode='wb') as ff:
                #     ff.write(contents)

            else:
                shutil.copytree(s, d)

                # while not os.path.isdir(d):
                #     time.sleep(0.001)

                # os.makedirs(d)
                # for root, sdirs, files, in os.walk(s):
                #     libroot = root.split(s)[-1].lstrip('/')
                #
                #     for di in sdirs:
                #         os.makedirs(os.path.join(d, libroot, di))
                #
                #     for fi in files:
                #         ss = os.path.join(root, fi)
                #         dd = os.path.join(d, libroot, fi)
                #         with open(ss, mode='rb') as ff:
                #             contents = ff.read()
                #         contents = contents.replace(self.simid.encode(), self.circuit.id.encode())
                #         with open(dd, mode='wb') as ff:
                #             ff.write(contents)

        for local_item in local_lib_contents:
            local_alias = local_item.split('_')
            local_alias = '_'.join(local_alias[:4] + local_alias[5:])

            library_contents = os.listdir(library_path)
            if any([local_alias == '_'.join(item.split('_')[:4]+item.split('_')[5:]) for item in library_contents]):
                continue

            global_manager.global_lock().acquire()  # get a lock
            library_contents = os.listdir(library_path)
            if any([local_alias == '_'.join(item.split('_')[:4] + item.split('_')[5:]) for item in library_contents]):
                # somebody did it while you were waiting!!
                global_manager.global_lock().release()
                continue
            try:
                s = os.path.join(local_ship_db, local_item)
                d = os.path.join(library_path, local_item)
                cache_item(s, d)
                print('cached AHDL model: {}'.format(d))
            finally:
                global_manager.global_lock().release()

    def _load_relevant_lib_items(self):
        loaded = False

        local_ship_db = os.path.join(self.working_path, local_ship_db_relpath)
        # everything should already be copied, just have to fix names and contents:

        original_ids = list(set([item.split('_')[4] for item in os.listdir(local_ship_db)]))
        original_ids = [item for item in original_ids if len(item) == len(self.simid) and '_' not in item]

        # see what needs to be renamed

        for id in original_ids:

            pairs = []

            for root, sdirs, files in os.walk(local_ship_db):
                for sdir in sdirs:
                    if id in sdir:
                        pairs.append((
                            os.path.join(root, sdir),
                            os.path.join(root, sdir.replace(id, self.simid))
                        ))
                for ffile in files:
                    if id in ffile:
                        pairs.append((
                            os.path.join(root, ffile),
                            os.path.join(root, ffile.replace(id, self.simid))
                        ))

            if pairs:
                loaded = True

                # rename everything
                for s, d in pairs:
                    shutil.move(s, d)

                # now go through and fix contents
                for _, dd in pairs:
                    if os.path.isfile(dd):
                        with open(dd, mode='rb') as ff:
                            contents = ff.read()
                        contents = contents.replace(id.encode(), self.simid.encode())
                        with open(dd, mode='wb') as ff:
                            ff.write(contents)

                    else:
                        for root, _, files in os.walk(dd):
                            for ffile in files:
                                with open(os.path.join(root, ffile), mode='rb') as ff:
                                    contents = ff.read()
                                contents = contents.replace(id.encode(), self.simid.encode())
                                with open(os.path.join(root, ffile), mode='wb') as ff:
                                    ff.write(contents)

        return loaded
        # library_path = os.path.join(pyspectre_ckt_dir, self.circuit.id, local_ship_db_relpath)
        # if not os.path.isdir(library_path):
        #     return
        # library_contents = os.listdir(library_path)
        # relevent_lib_items = [item for item in library_contents if self.circuit.id in item]
        #
        # if relevent_lib_items:
        #     # found relevant library item... copy it into the local library
        #     for item in relevent_lib_items:
        #         newname = item.replace(id_rep, path_rep)
        #         s = os.path.join(pyspectre_ahdl_lib, item)
        #         d = os.path.join(local_ship_db, newname)
        #         if os.path.isdir(s):
        #             shutil.copytree(s, d)
        #         elif os.path.isfile(s):
        #             shutil.copy(s, d)
        #
        #             if any([re.match(pattern, item) for pattern in
        #                     _get_patterns_for_internal_replacement(id_rep)]):
        #                 contents = open(d, mode='rb').read()
        #                 contents = contents.replace(id_rep.encode(), path_rep.encode())
        #                 open(d, mode='wb').write(contents)

    def _read_file(self, filename, signames='all'):
        if not os.path.isfile(filename):
            filename = os.path.join(self.working_path, 'psf', filename)

        if isinstance(signames, str) and not signames == 'all':
            signames = [signames]

        if self.reader == 'libpsf':
            try:
                d = libpsf.PSFDataSet(filename)
            except IOError as e:
                print(
                    'ERROR: libpsf couldn\'t open file: '
                    '{}. See spectre simulation directory: {}'.format(filename, self.working_path)
                )
                raise e

            if isinstance(signames, str) and signames == 'all':
                return d.get_sweep_values().flatten(), np.array([d.get_signal(signame)
                                                                 for signame in d.get_signal_names()]).T
            else:
                return d.get_sweep_values().flatten(), np.array([d.get_signal(signame)
                                                                 for signame in signames]).T

        else:
            raise NotImplementedError

    def _resolve_analysis_type(self):
        self.analysis_type = self.interface.listanalysis()[self.analysis_name]
        # self.analysis = eval('Analyses.' + self.analysis_type.title() + '()')

    def _extend_pwl_defs_in_netlist(self):
        """
        create an alterate version of the netlist with reservation in the wave="*" assignment
        for a pwl signal as long as was requested
        :return:
        """

        modfile_suffix = '_PYSPECTRE_ALTERED'
        modfile_prefix = ''

        if self.pwl_inputs is None:
            return
        targets = [item.cname for item in self.pwl_inputs]
        target_names = [item.split('.')[-1] for item in targets]

        netlist = open(self.spicefilepath, mode='r').read().splitlines()

        # go through and find definitions:
        def_lines = {}
        context = []
        for line_no, line in enumerate(netlist):
            line = line.lstrip()
            if line.startswith('//'):
                continue
            toks = line.split()
            if len(toks) < 1:
                continue
            anchor = toks[0]
            if anchor == 'subckt':
                context.append(toks[1])
                continue
            if anchor == 'ends':
                context.pop()
                continue
            if anchor in target_names:
                this_guys_name = '.'.join(context) + '.' + anchor if context else anchor
                if this_guys_name in targets:
                    stopline = line_no
                    while re.match(r'.*\\\s*$', netlist[stopline]):
                        stopline += 1
                    def_lines.update({this_guys_name: {'PwlInput': target_names.index(anchor),
                                                       'range': range(line_no, stopline + 1)}})

        assert all([item in def_lines for item in targets])

        file_pattern = r'^.*(\s+file=\".+?\")'
        wave_pattern = r'^.*(\s+wave=[\[\"].+?[\]\"])'

        for name, patch_dict in def_lines.items():

            fulldef = ' '.join([netlist[ll].replace('\\', '').strip() for ll in patch_dict['range']])

            # remove all 'file' definitions

            if re.match(file_pattern, fulldef):
                fulldef = ' '.join(
                    [item.strip() for item in fulldef.split(re.match(file_pattern, fulldef).groups()[0])]).strip()

            # check and extend existing 'wave' definition
            if re.match(wave_pattern, fulldef):
                oldwave = re.match(wave_pattern, fulldef).groups()[0].split('=')[1].strip().strip('\"[]').split()
                fulldef = ' '.join(
                    [item.strip() for item in fulldef.split(re.match(wave_pattern, fulldef).groups()[0])]).strip()
                old_time = np.stack([float(t) for t in oldwave[::2]])
                old_vals = np.stack([float(v) for v in oldwave[1::2]])

                if old_time.size >= self.pwl_inputs[target_names.index(name)].max_n_points:
                    continue
                else:
                    aug_len = self.pwl_inputs[target_names.index(name)].max_n_points - old_time.size
                    ts = np.max(np.diff(old_time))
                    new_time = np.concatenate(
                        [old_time, np.linspace(old_time[-1] + ts, old_time[-1] + aug_len * ts, aug_len)])
                    new_vals = np.concatenate([old_vals, np.zeros(aug_len)])

                    newval_str = ' '.join(
                        [' '.join([float_fmt_string] * 2).format(new_time[i], new_vals[i]) for i in range(aug_len)])

            else:
                # create brand-new 'wave' definition
                aug_len = self.pwl_inputs[patch_dict['PwlInput']].max_n_points
                ts = 1
                new_time = np.linspace(0, (aug_len - 1) * ts, aug_len)
                new_vals = np.random.randn(aug_len)

                newval_str = ' '.join(
                    [' '.join([float_fmt_string] * 2).format(new_time[i], new_vals[i]) for i in range(aug_len)])

            fulldef = ' '.join([fulldef, 'wave=[' + newval_str + ']'])
            fulldef = textwrap(fulldef, break_long_words=False, break_on_hyphens=False, subsequent_indent='    ')
            fulldef[:-1] = [item + ' \\' for item in fulldef[:-1]]  ## fwd slash on ends of lines
            # fulldef[1:] = ['+ ' + item for item in fulldef[1:]]     ## '+' at beginning of lines
            def_lines[name].update({'def': [ll + '\n' for ll in fulldef]})

        # all fixed up!
        # shutil.copy(self.spicefilepath, self.spicefilepath+'_bk')
        patchlist = [item['range'] for item in def_lines.values()]

        orig_parts = os.path.split(self.spicefilepath)
        orig_root = orig_parts[0]
        orig_base = orig_parts[1].split(os.path.extsep)
        orig_ext = os.path.extsep + orig_base[-1]
        orig_base = os.path.extsep.join(orig_base[:-1])

        target_file = os.path.join(orig_root, modfile_prefix+orig_base+modfile_suffix+orig_ext)
        with open(target_file, mode='w') as outfile:
            for line_no, line in enumerate(netlist):
                if any([line_no in r for r in patchlist]):
                    patch = [v for v in def_lines.values() if line_no in v['range']][0]
                    if line_no + 1 not in patch['range']:
                        outfile.writelines(patch['def'])
                        del patchlist[patchlist.index(patch['range'])]
                else:
                    outfile.write(line + '\n')

        self.spicefilepath = target_file

    def getparams(self, obj, paramnames):
        if isinstance(paramnames, str):
            paramnames = [paramnames]
        paramlist = self.interface.listparams(obj)
        if not all([name in paramlist for name in paramnames]):
            missing = [name not in paramlist for name in paramnames]
            print('coudln\'t find parameters named:{} in object:{}'.format(missing, obj))
            raise PyspectreUnresolvableError
        return {name: self.interface.getparametervalue(obj, name) for name in paramnames}

    def setparams(self, obj, paramdict):
        assert isinstance(paramdict, dict)
        for name, value in paramdict.items():
            self.interface.setparametervalue(obj, name, value)

    def setpwl(self, name, time_vec, volts):

        if isinstance(name, PwlInput):
            name = name.cname

        assert len(time_vec.shape) <= 2
        if len(time_vec.shape) == 2:
            assert np.min(time_vec.shape) == 1
            time_vec = time_vec.squeeze()

        assert len(volts.shape) <= 2
        if len(volts.shape) == 2:
            assert np.min(volts.shape) == 1
            volts = volts.squeeze()

        assert time_vec.size == volts.size, 'incompatible time vector'

        as_text = '[' + format_vect(np.stack([time_vec, volts])) + ']'

        self.setparams(name, {'wave': as_text})

    def close(self):
        self._cleanup()

    def runanalysis(self, analysis_name=None):
        if analysis_name is None:
            if self.analysis_name is not None:
                analysis_name = self.analysis_name
            else:
                raise ValueError('must specify an analysis name')

        try:
            self.interface.runanalysis(analysis_name)

        except KeyboardInterrupt as e:
            raise e

        except InterruptedError as e:
            raise e

        except Exception as e:
            print('analysis failed')
            if self.logging:
                shutil.copytree(self.working_path, os.path.join(pyspectre_path, 'crashed', self.simid))
                print('copied simdir to {} for debugging'.format(os.path.join(pyspectre_path, 'crashed', self.simid)))
            self._cleanup()
            raise e

    def setresultsdir(self, relpath=None):
        if relpath is None:
            if self.results_dir is not None:
                relpath = self.results_dir
            else:
                raise ValueError('no path specified')

        self.interface.setresultsdir(relpath)

    def readstatefile(self, filename=None):
        statedata = {}
        with open(os.path.join(self.working_path, filename)) as rdata:
            for line in rdata:
                if line.startswith('#'):
                    continue
                else:
                    tok = line.split()
                    statedata.update({tok[0]: float(tok[1])})

        return statedata

    def readdcfile(self):
        return self.readstatefile(self.analysis.dcfile)

    def readtranfile(self):
        return self.readstatefile(self.analysis.tranfile)

    def init(self, *args, **kwargs):
        """ use this for one-time calls which will persist forever"""
        self.analysis.init(self, *args, **kwargs)
        return self.analysis.dcfile

    def reset(self, *args, **kwargs):
        """ use this for one-time prep calls which will persist between analyses"""
        self.analysis.reset(self, *args, **kwargs)
        return self.analysis.tranfile

    def step(self, *args, **kwargs):
        """ use this for calls which have to be made before every step"""
        self.analysis.step(self, *args, **kwargs)
        return self.analysis.tranfile

    def get_data(self, signames=None):
        if isinstance(signames, (str)):
            signames = [signames]
        t, x = self.analysis.interpret_data(self, signames)
        return t, x

    def read_results(self, signal_names='all', analysis_name=None, analysis_type=None):
        if analysis_name is None:
            if self.analysis_name is not None:
                analysis_name = self.analysis_name
            else:
                raise ValueError('must specify analysis name')

        if analysis_type is None:
            if self.analysis_type is not None:
                analysis_type = self.analysis_type
            else:
                raise ValueError('must specify analysis type')

        return self._read_file(
            os.path.join(self.results_dir, '.'.join([analysis_name, analysis_type])),
            signames=signal_names
        )

    def run_batch(self, list_of_stimuli, list_of_times, signal_names='all', resample=False, verbosity=None):

        if verbosity is None:
            if self.verbosity is not None:
                verbosity = self.verbosity
            else:
                verbosity = 1

        if len(list_of_times.shape) == 1:
            list_of_times = np.expand_dims(list_of_times, 0)
        if len(list_of_times.shape) > 2:
            raise ValueError('Time vector must be 1 or 2 dimensional')
        if 1 in list_of_times.shape:
            list_of_times = list_of_times.reshape(1, -1)
            list_of_times = np.broadcast_to(list_of_times, (list_of_stimuli.shape[0], list_of_times.shape[1]))

        assert list_of_stimuli.shape[0] == list_of_times.shape[0]
        assert list_of_stimuli.shape[1] == list_of_times.shape[1]

        print('starting batch of {} simulations.'.format(len(list_of_stimuli))) if verbosity >= 1 else None
        restimes = []
        ressignals = []
        otime = time.time()
        for time_vec, stimulus in zip(list_of_times, list_of_stimuli):

            ootime = time.time()
            for i, input in enumerate(self.pwl_inputs):
                self.setpwl(input, time_vec=time_vec, volts=stimulus[:, i])

            self.setparams(self.analysis_name, {'stop': time_vec[-1]})
            self.runanalysis()

            restime, ressigs = self.read_results(signal_names=signal_names)

            if resample:
                n_blocks = time_vec.size

                t_len = int(
                    (np.floor(restime.size/n_blocks)
                     if not np.floor(restime.size/n_blocks) % 2 else np.floor(restime.size/n_blocks)-1
                     ) * n_blocks)

                if t_len == 0:
                    t_len = int(
                        (np.ceil(restime.size / n_blocks)
                         if not np.ceil(restime.size / n_blocks) % 2 else np.ceil(restime.size / n_blocks) + 1
                         ) * n_blocks)

                reg_samps = interp1d(
                    x=restime,
                    y=ressigs,
                    kind='linear',
                    axis=0,
                    bounds_error=False,
                    fill_value=(ressigs[0, :], ressigs[-1, :]),
                    assume_sorted=True,
                )(np.linspace(time_vec[0], time_vec[-1], t_len, endpoint=True))

                block_size = int(reg_samps.shape[0]/n_blocks)
                split_inds = np.cumsum(
                    np.stack([block_size]*n_blocks),
                )[:-1].astype(int)
                new_sigs = np.stack([
                    np.mean(item, axis=0) for item in np.array_split(reg_samps, split_inds, axis=0)
                ])

                ressigs = new_sigs
                restime = time_vec

            print('simulation took {:4.1f} ms'.format((time.time() - ootime) * 1000)) if verbosity >= 2 else None
            ressignals.append(ressigs)
            restimes.append(restime)

        r_time = time.time() - otime
        results = restimes, ressignals

        if r_time < 1:
            print('[pyspectre] batch took {:4.1f} ms'.format(r_time*1000)) if verbosity >= 1 else None
        elif 60 > r_time >= 1:
            print('[pyspectre] batch took {:4.1f} s'.format(r_time)) if verbosity >= 1 else None
        elif r_time >= 60:
            print('[pyspectre] batch took {:4.1f} m'.format(r_time/60)) if verbosity >= 1 else None

        return results

